@@ -1,50 +1,52 @@
 import asyncio
 import json
 import logging
+import re
 import sys
 import uuid
-from datetime import datetime, timezone, timedelta
+from datetime import datetime, timedelta
+from itertools import cycle
 from timeit import default_timer as timer
-from typing import Optional, Set, Tuple, List
+from typing import List, Set, Tuple, Union, Optional
 
-import beem
 import rfc3987
-from beem.account import Account
-from beem.exceptions import AccountDoesNotExistsException, MissingKeyError
-from beemapi.exceptions import UnhandledRPCError
+from lighthive.datastructures import Operation
+from lighthive.exceptions import RPCNodeException
+from lighthive.node_picker import compare_nodes
 
+from podping_hivewriter import __version__ as podping_hivewriter_version
 from podping_hivewriter.async_context import AsyncContext
+from podping_hivewriter.async_wrapper import sync_to_async
 from podping_hivewriter.constants import (
-    STARTUP_FAILED_HIVE_API_ERROR_EXIT_CODE,
+    HIVE_CUSTOM_OP_DATA_MAX_LENGTH,
     STARTUP_FAILED_INVALID_POSTING_KEY_EXIT_CODE,
     STARTUP_FAILED_UNKNOWN_EXIT_CODE,
     STARTUP_OPERATION_ID,
-    CURRENT_PODPING_VERSION,
-    HIVE_CUSTOM_OP_DATA_MAX_LENGTH,
 )
-from podping_hivewriter.exceptions import PodpingCustomJsonPayloadExceeded
-from podping_hivewriter.hive_wrapper import HiveWrapper
+from podping_hivewriter.exceptions import (
+    PodpingCustomJsonPayloadExceeded,
+    TooManyCustomJsonsPerBlock,
+)
+from podping_hivewriter.hive import get_allowed_accounts, get_client
+from podping_hivewriter.models.hive_operation_id import HiveOperationId
 from podping_hivewriter.models.iri_batch import IRIBatch
+from podping_hivewriter.models.medium import Medium
+from podping_hivewriter.models.podping import Podping
+from podping_hivewriter.models.reason import Reason
 from podping_hivewriter.podping_settings_manager import PodpingSettingsManager
 
 
-def utc_date_str() -> str:
-    return datetime.utcnow().replace(tzinfo=timezone.utc).isoformat()
-
-
-def size_of_dict_as_json(payload: dict):
-    return len(json.dumps(payload, separators=(",", ":")).encode("UTF-8"))
-
-
 class PodpingHivewriter(AsyncContext):
     def __init__(
         self,
         server_account: str,
         posting_keys: List[str],
         settings_manager: PodpingSettingsManager,
+        medium: Medium = Medium.podcast,
+        reason: Reason = Reason.update,
         listen_ip: str = "127.0.0.1",
         listen_port: int = 9999,
-        operation_id="podping",
+        operation_id="pp",
         resource_test=True,
         dry_run=False,
         daemon=True,
@@ -55,6 +57,8 @@ class PodpingHivewriter(AsyncContext):
         self.server_account: str = server_account
         self.required_posting_auths = [self.server_account]
         self.settings_manager = settings_manager
+        self.medium = medium
+        self.reason = reason
         self.listen_ip = listen_ip
         self.listen_port = listen_port
         self.posting_keys: List[str] = posting_keys
@@ -64,8 +68,14 @@ class PodpingHivewriter(AsyncContext):
         self.daemon: bool = daemon
         self.status: bool = status
 
-        self.hive_wrapper = HiveWrapper(
-            posting_keys, settings_manager, dry_run=dry_run, daemon=daemon
+        self.lighthive_client = get_client(
+            posting_keys=posting_keys,
+            loglevel=logging.ERROR,
+            automatic_node_selection=False,  # TODO: File upstream lighthive bug because it runs asyncio in a new loop
+        )
+
+        self._async_hive_broadcast = sync_to_async(
+            self.lighthive_client.broadcast, thread_sensitive=False
         )
 
         self.total_iris_recv = 0
@@ -87,12 +97,9 @@ class PodpingHivewriter(AsyncContext):
     async def _startup(self):
 
         try:
-            hive = await self.hive_wrapper.get_hive()
-            account = Account(self.server_account, blockchain_instance=hive, lazy=True)
             settings = await self.settings_manager.get_settings()
-
             allowed = get_allowed_accounts(
-                settings.main_nodes, settings.control_account
+                self.lighthive_client, settings.control_account
             )
             # TODO: Should we periodically check if the account is allowed
             #  and shut down if not?
@@ -100,19 +107,14 @@ class PodpingHivewriter(AsyncContext):
                 logging.error(
                     f"Account @{self.server_account} not authorised to send Podpings"
                 )
-        except AccountDoesNotExistsException:
-            logging.error(
-                f"Hive account @{self.server_account} does not exist, "
-                f"check ENV vars and try again",
-                exc_info=True,
-            )
-            raise
-        except Exception:
-            logging.error("Unknown error occurred", exc_info=True)
-            raise
+
+        except Exception as ex:
+            logging.error(f"Unknown error occurred: {ex}", exc_info=True)
+            raise ex
 
         if self.resource_test and not self.dry_run:
-            await self.test_hive_resources(account, hive)
+            await self.test_hive_resources()
+            await self.automatic_node_selection()
 
         logging.info(f"Hive account: @{self.server_account}")
 
@@ -125,7 +127,16 @@ class PodpingHivewriter(AsyncContext):
 
         self._startup_done = True
 
-    async def test_hive_resources(self, account, hive):
+    async def automatic_node_selection(self) -> None:
+        """Use the automatic async feature to find the fastest API"""
+        self.lighthive_client._node_list = await compare_nodes(
+            nodes=self.lighthive_client.nodes, logger=self.lighthive_client.logger
+        )
+        self.lighthive_client.node_list = cycle(self.lighthive_client._node_list)
+        self.lighthive_client.next_node()
+        logging.info(f"Lighthive Fastest: {self.lighthive_client.current_node}")
+
+    async def test_hive_resources(self):
         logging.info(
             "Podping startup sequence initiated, please stand by, "
             "full bozo checks in operation..."
@@ -133,64 +144,62 @@ class PodpingHivewriter(AsyncContext):
 
         # noinspection PyBroadException
         try:  # Now post two custom json to test.
-            manabar = account.get_rc_manabar()
+            account = self.lighthive_client.account(self.server_account)
+            manabar = account.get_resource_credit_info()
+
             logging.info(
                 f"Testing Account Resource Credits"
-                f' - before {manabar.get("current_pct"):.2f}%'
+                f' - before {manabar.get("last_mana_percent"):.2f}%'
             )
+            rc = self.lighthive_client.rc()
 
-            # TODO: See if anything depends on USE_TEST_NODE before removal
             custom_json = {
                 "server_account": self.server_account,
-                "USE_TEST_NODE": False,
                 "message": "Podping startup initiated",
                 "uuid": str(uuid.uuid4()),
-                "hive": repr(hive),
+                "hive": str(self.lighthive_client.current_node),
             }
 
-            await self.send_notification(custom_json, STARTUP_OPERATION_ID)
+            startup_hive_operation_id = self.operation_id + STARTUP_OPERATION_ID
 
-            logging.info("Testing Account Resource Credits.... 5s")
-            settings = await self.settings_manager.get_settings()
-            await asyncio.sleep(settings.hive_operation_period)
-            manabar_after = account.get_rc_manabar()
+            op, size_of_json = await self.construct_operation(
+                custom_json, startup_hive_operation_id
+            )
+            rc_cost = rc.get_cost(op)
+            percent_after = (
+                100
+                * (manabar.get("last_mana") - (1e6 * rc_cost * 100))
+                / manabar["max_mana"]
+            )
+            percent_drop = manabar.get("last_mana_percent") - percent_after
+            capacity = (100 / percent_drop) * 100
             logging.info(
-                f"Testing Account Resource Credits"
-                f' - after {manabar_after.get("current_pct"):.2f}%'
+                f"Calculating Account Resource Credits "
+                f"for 100 pings: {percent_drop:.2f}% | "
+                f"Capacity: {capacity:,.0f}"
             )
-            cost = manabar.get("current_mana") - manabar_after.get("current_mana")
-            if cost == 0:  # skip this test if we're going to get ZeroDivision
-                capacity = 1000000
-            else:
-                capacity = manabar_after.get("current_mana") / cost
-            logging.info(f"Capacity for further podpings : {capacity:.1f}")
 
-            custom_json["v"] = CURRENT_PODPING_VERSION
-            custom_json["capacity"] = f"{capacity:.1f}"
+            custom_json["v"] = podping_hivewriter_version
+            custom_json["capacity"] = f"{capacity:,.0f}"
             custom_json["message"] = "Podping startup complete"
-            custom_json["hive"] = repr(hive)
+            custom_json["hive"] = str(self.lighthive_client.current_node)
 
-            await self.send_notification(custom_json, STARTUP_OPERATION_ID)
+            await self.send_notification(custom_json, startup_hive_operation_id)
 
             logging.info("Startup of Podping status: SUCCESS! Hit the BOOST Button.")
 
-        except MissingKeyError as _:
+        except ValueError as ex:
+            if str(ex) == "Error loading Base58 object":
+                logging.error(
+                    f"Startup of Podping status: FAILED!  {ex}",
+                    exc_info=True,
+                )
+                logging.error("Exiting")
+                sys.exit(STARTUP_FAILED_INVALID_POSTING_KEY_EXIT_CODE)
+
+        except Exception as ex:
             logging.error(
-                "Startup of Podping status: FAILED!  Invalid posting key",
-                exc_info=True,
-            )
-            logging.error("Exiting")
-            sys.exit(STARTUP_FAILED_INVALID_POSTING_KEY_EXIT_CODE)
-        except UnhandledRPCError as _:
-            logging.error(
-                "Startup of Podping status: FAILED!  API error",
-                exc_info=True,
-            )
-            logging.info("Exiting")
-            sys.exit(STARTUP_FAILED_HIVE_API_ERROR_EXIT_CODE)
-        except Exception as _:
-            logging.error(
-                "Startup of Podping status: FAILED!  Unknown error",
+                f"Startup of Podping status: FAILED!  {ex}",
                 exc_info=True,
             )
             logging.error("Exiting")
@@ -219,23 +228,27 @@ class PodpingHivewriter(AsyncContext):
                 iri_batch = await self.iri_batch_queue.get()
 
                 start = timer()
-                trx_id, failure_count = await self.failure_retry(iri_batch.iri_set)
+                failure_count = await self.failure_retry(
+                    iri_batch.iri_set, medium=self.medium, reason=self.reason
+                )
                 duration = timer() - start
 
                 self.iri_batch_queue.task_done()
                 async with self._iris_in_flight_lock:
                     self._iris_in_flight -= len(iri_batch.iri_set)
 
-                hive = await self.hive_wrapper.get_hive()
-                last_node = hive.data["last_node"]
+                last_node = self.lighthive_client.current_node
                 logging.info(
-                    f"Batch send time: {duration:0.2f} - trx_id: {trx_id} - "
-                    f"Failures: {failure_count} - IRI batch_id {iri_batch.batch_id} - "
-                    f"IRIs in batch: {len(iri_batch.iri_set)} - "
+                    f"Batch send time: {duration:0.2f} | "
+                    f"Failures: {failure_count} - IRI batch_id {iri_batch.batch_id} | "
+                    f"IRIs in batch: {len(iri_batch.iri_set)} | "
                     f"last_node: {last_node}"
                 )
             except asyncio.CancelledError:
                 raise
+            except Exception as ex:
+                logging.error(f"{ex} occurred", exc_info=True)
+                raise
 
     async def _iri_batch_loop(self):
         async def get_from_queue():
@@ -339,72 +352,117 @@ class PodpingHivewriter(AsyncContext):
     async def output_hive_status(self) -> None:
         """Output the name of the current hive node
         on a regular basis"""
-        up_time = timedelta(seconds=timer() - self.startup_time)
-
-        hive = await self.hive_wrapper.get_hive()
-        last_node = hive.data["last_node"]
+        up_time = timedelta(seconds=int(timer() - self.startup_time))
+        await self.automatic_node_selection()
+        last_node = self.lighthive_client.current_node
         logging.info(
-            f"Status - Uptime: {up_time} - "
-            f"IRIs Received: {self.total_iris_recv} - "
-            f"IRIs Deduped: {self.total_iris_recv_deduped} - "
-            f"IRIs Sent: {self.total_iris_sent} - "
+            f"Status - Uptime: {up_time} | "
+            f"IRIs Received: {self.total_iris_recv} | "
+            f"IRIs Deduped: {self.total_iris_recv_deduped} | "
+            f"IRIs Sent: {self.total_iris_sent} | "
             f"last_node: {last_node}"
         )
 
+    async def construct_operation(
+        self, payload: dict, hive_operation_id: Union[HiveOperationId, str]
+    ) -> Tuple[Operation, int]:
+        """Builed the operation for the blockchain"""
+        payload_json = json.dumps(payload, separators=(",", ":"))
+        size_of_json = len(payload_json)
+        if size_of_json > HIVE_CUSTOM_OP_DATA_MAX_LENGTH:
+            raise PodpingCustomJsonPayloadExceeded("Max custom_json payload exceeded")
+
+        op = Operation(
+            "custom_json",
+            {
+                "required_auths": [],
+                "required_posting_auths": self.required_posting_auths,
+                "id": str(hive_operation_id),
+                "json": payload_json,
+            },
+        )
+        return op, size_of_json
+
     async def send_notification(
-        self, payload: dict, operation_id: Optional[str] = None
-    ) -> str:
+        self, payload: dict, hive_operation_id: Union[HiveOperationId, str]
+    ) -> None:
+        """Build and send an operation to the blockchain"""
         try:
-            size_of_json = size_of_dict_as_json(payload)
-            if size_of_json > HIVE_CUSTOM_OP_DATA_MAX_LENGTH:
-                raise PodpingCustomJsonPayloadExceeded(
-                    "Max custom_json payload exceeded"
-                )
-            tx = await self.hive_wrapper.custom_json(
-                operation_id or self.operation_id, payload, self.required_posting_auths
+            op, size_of_json = await self.construct_operation(
+                payload, hive_operation_id
             )
+            # if you want to FORCE the error condition for >5 operations
+            # in one block, uncomment this line.
+            # op = [op] * 6
 
-            tx_id = tx["trx_id"]
+            # Use asynchronous broadcast but means we don't get back tx, kinder to
+            # API servers
+            await self._async_hive_broadcast(op=op, dry_run=self.dry_run)
 
-            logging.info(f"Transaction sent: {tx_id} - JSON size: {size_of_json}")
+            logging.info(f"Lighthive Node: {self.lighthive_client.current_node}")
+            logging.info(f"JSON size: {size_of_json}")
 
-            return tx_id
+        except RPCNodeException as ex:
+            logging.error(f"send_notification error: {ex}")
+            if re.match(
+                r"plugin exception.*custom json.*", ex.raw_body["error"]["message"]
+            ):
+                self.lighthive_client.next_node()
+                raise TooManyCustomJsonsPerBlock()
+            raise ex
 
-        except MissingKeyError:
-            logging.error(f"The provided key for @{self.server_account} is not valid")
-            raise
+        except PodpingCustomJsonPayloadExceeded as ex:
+            raise ex
 
-    async def send_notification_iri(self, iri: str, reason="feed_update") -> str:
-        payload = {
-            "version": CURRENT_PODPING_VERSION,
-            "num_urls": 1,
-            "reason": reason,
-            "urls": [iri],
-        }
-        return await self.send_notification(payload)
+        except Exception as ex:
+            logging.error(f"{ex}")
+            raise ex
 
-    async def send_notification_iris(self, iris: Set[str], reason="feed_update") -> str:
+    async def send_notification_iri(
+        self,
+        iri: str,
+        medium: Optional[Medium],
+        reason: Optional[Reason],
+    ) -> None:
+        payload = Podping(
+            medium=medium or self.medium, reason=reason or self.reason, iris=[iri]
+        )
+
+        hive_operation_id = HiveOperationId(self.operation_id, medium, reason)
+
+        await self.send_notification(payload.dict(), hive_operation_id)
+
+        self.total_iris_sent += 1
+
+    async def send_notification_iris(
+        self,
+        iris: Set[str],
+        medium: Optional[Medium],
+        reason: Optional[Reason],
+    ) -> None:
         num_iris = len(iris)
-        payload = {
-            "version": CURRENT_PODPING_VERSION,
-            "num_urls": num_iris,
-            "reason": reason,
-            "urls": list(iris),
-        }
+        payload = Podping(
+            medium=medium or self.medium, reason=reason or self.reason, iris=list(iris)
+        )
 
-        tx_id = await self.send_notification(payload)
+        hive_operation_id = HiveOperationId(self.operation_id, medium, reason)
+
+        await self.send_notification(payload.dict(), hive_operation_id)
 
         self.total_iris_sent += num_iris
 
-        return tx_id
-
-    async def failure_retry(self, iri_set: Set[str]) -> Tuple[str, int]:
+    async def failure_retry(
+        self,
+        iri_set: Set[str],
+        medium: Optional[Medium],
+        reason: Optional[Reason],
+    ) -> int:
         await self.wait_startup()
         failure_count = 0
 
         while True:
-            # Sleep a maximum of 5 minutes, 2 additional seconds for every retry
-            sleep_time = min(failure_count * 2, 300)
+            # Sleep a maximum of 5 minutes, 3 additional seconds for every retry
+            sleep_time = min(failure_count * 3, 300)
             if failure_count > 0:
                 logging.warning(f"Waiting {sleep_time}s before retry")
                 await asyncio.sleep(sleep_time)
@@ -415,34 +473,35 @@ class PodpingHivewriter(AsyncContext):
                 logging.info(f"Received {len(iri_set)} IRIs")
 
             try:
-                trx_id = await self.send_notification_iris(iris=iri_set)
+                await self.send_notification_iris(
+                    iris=iri_set,
+                    medium=medium or self.medium,
+                    reason=reason or self.reason,
+                )
                 if failure_count > 0:
                     logging.info(
                         f"FAILURE CLEARED after {failure_count} retries, {sleep_time}s"
                     )
-                return trx_id, failure_count
-            except Exception:
+                return failure_count
+            except RPCNodeException as ex:
+                logging.warning(f"{ex}")
                 logging.warning(f"Failed to send {len(iri_set)} IRIs")
+                if ex.raw_body["error"]["data"]["name"] == "tx_missing_posting_auth":
+                    for iri in iri_set:
+                        logging.error(iri)
+                    logging.error(
+                        f"Terminating: exit code: "
+                        f"{STARTUP_FAILED_INVALID_POSTING_KEY_EXIT_CODE}"
+                    )
+                    sys.exit(STARTUP_FAILED_INVALID_POSTING_KEY_EXIT_CODE)
+
+            except Exception as ex:
+                logging.warning(f"Failed to send {len(iri_set)} IRIs")
+                logging.warning(f"{ex}")
                 if logging.DEBUG >= logging.root.level:
                     for iri in iri_set:
                         logging.debug(iri)
-                await self.hive_wrapper.rotate_nodes()
 
+            finally:
+                self.lighthive_client.next_node()
                 failure_count += 1
-
-
-def get_allowed_accounts(
-    nodes: Tuple[str, ...], account_name: str = "podping"
-) -> Set[str]:
-    """get a list of all accounts allowed to post by acc_name (podping)
-    and only react to these accounts"""
-
-    try:
-        hive = beem.Hive(node=nodes)
-        master_account = Account(account_name, blockchain_instance=hive, lazy=True)
-        return set(master_account.get_following())
-    except Exception:
-        logging.error(
-            f"Allowed Account: {account_name} - Failure on Node: {nodes[0]}",
-            exc_info=True,
-        )
